/// Josh's User Interface Library
/// A C++20 Library for creating, styling, and rendering of a UI/UX widgets.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file TextInputForm.hpp
/// @desc A box that accepts user keyboard input.
/// @edit 2024-08-02

#pragma once

#include <set>
#include <JUI/Mixins/Clickable.hpp>

#include "TextRect.hpp"

namespace JUI {

    // TODO: Text Selection
    // TODO: Support Copy
    // TODO: Support Paste
    // TODO: Support Cut
    // TODO: Support insert at cursor
    // TODO: Simulate key repeat.

    class TextInputForm : public TextRect, public Clickable {
    public:
        Event<> OnSelect;
        Event<> OnDeselect;
        Event<std::string> OnReturn;
        TextInputForm();
        explicit TextInputForm(Widget* parent);
        void Update(float elapsed) override;
        void InnerDraw() override;
        void Draw() override;

        void MoveCursorLeft();

        void MoveCursorRight();

        /// Returns the maximum position of the cursor, which is determined by the input buffer's length.
        [[nodiscard]] unsigned int CursorMaxPosition() const;

        void SetCursorPosition(unsigned int pos);

        /// Fires the input event, clears the input buffer, and sets the state to be ready for further input.
        void SendInput(bool clear_input);

        /// Copy the contents of the input form, or the selection if active, into the the system's clipboard.
        void Copy();
        /// Paste the contents of the system's clipboard 
        void Paste();

        void Cut();

        void Backspace();

        void Delete();

        void PushStringToCurrentPlaceInInputBuffer(const std::string &snippet);

        void PushKeyToCurrentPlaceInInputBuffer(const Key &key);

        bool ObserveKeyInput(Key key, bool pressed) override;
        bool ObserveMouseInput(MouseButton btn, bool pressed) override;
        bool ObserveMouseMovement(const Vector2 &latest_known_pos) override;
        [[nodiscard]] std::string GetAutocompleteText() const;
        void SetAutoCompleteText(const std::string& text);
        [[nodiscard]] Color4 GetAutocompleteTextColor() const;
        void SetAutocompleteTextColor(const Color4& color);
        [[nodiscard]] bool HideAutocompleteOnSelect() const;
        void SetHideAutocompleteOnSelect(bool hide);
        [[nodiscard]] bool AutocompleteTextEnabled() const;
        void SetAutocompleteTextEnabled(bool enabled);

        [[nodiscard]] std::string InputBuffer() const;
        void SetInputBuffer(const std::string& value);
        void ClearInputBuffer();


        [[nodiscard]] bool HasFocus() const;
        void SetFocused(bool focused);

        // TODO: Implement procedure to allow API consumer to validate the input, **before** clearing the buffer.

        [[nodiscard]] bool ClearTextOnReturn() const;
        void ClearTextOnReturn(bool value);

        [[nodiscard]] bool DropFocusOnReturn() const;
        void DropFocusOnReturn(bool value);

        void GrabFocus();
        void DropFocus();

        [[nodiscard]] std::set<std::string> GetBlacklist() const;
        void SetBlacklist(const std::set<std::string>& value);
        void AddToBlacklist(const std::string& value);

        // TODO: Implement selection of part of input text.
        Color4 GetSelectionColor() const;
        void SetSelectionColor(const Color4& color);
        bool HasSelection() const;
        std::string GetSelectedText() const;
    protected:
        bool clear_text_on_return = true;
        bool drop_focus_on_return = false;
        bool focused = false;
        bool selection_enabled;
        bool selection_active;
        int selection_start_index;
        int selection_end_index;
        unsigned int cursor_position = 0;
        std::string input_buffer;
        float cursor_blink_time = 0.f;
        Color4 autocomplete_color = Style::InputForm::AutocompleteTextColor;
        std::string autocomplete_text = "Hello World";
        bool hide_autocomplete_on_select = true;
        bool autocomplete_text_enabled = true;
        std::set<std::string> blacklist;
    private:
    };
}